#include <assert.h>
#include <isl/set.h>
#include <isl/stream.h>

int main(int argc, char *argv[])
{
	int i;
	int params;
	FILE *f;
	isl_ctx *ctx;
	isl_set *set;
	struct isl_stream *s;
	struct isl_token *token;
	unsigned nparam;

	ctx = isl_ctx_alloc();
	assert(argc == 2);
	f = fopen(argv[1], "r");
	assert(f);
	s = isl_stream_new_file(ctx, f);
	isl_stream_eat(s, ISL_TOKEN_IDENT); /* C */

	set = isl_stream_read_set(s);

	nparam = isl_set_dim(set, isl_dim_set);
	set = isl_set_move_dims(set, isl_dim_param, 0, isl_dim_set, 0, nparam);

	token = isl_stream_next_token(s);
	assert(token);
	assert(token->type == ISL_TOKEN_VALUE);
	params = isl_int_is_one(token->u.v);
	isl_token_free(token);

	if (params) {
		for (i = 0; i < nparam; ++i) {
			token = isl_stream_next_token(s);
			assert(token->type == ISL_TOKEN_IDENT);
			set = isl_set_set_dim_name(set, isl_dim_param, i,
							token->u.s);
			isl_token_free(token);
		}
	} else {
		char name[] = "M";
		for (i = 0; i < nparam; ++i) {
			set = isl_set_set_dim_name(set, isl_dim_param, i, name);
			name[0]++;
		}
	}

	isl_set_print(set, stdout, 0, ISL_FORMAT_ISL);

	isl_set_free(set);
	isl_stream_free(s);
	isl_ctx_free(ctx);
	fclose(f);

	return 0;
}
