#include "suif_inc.h"

enum condition_tree_type {
	ctt_invalid,
	ctt_and,
	ctt_or,
	ctt_ge,
	ctt_eq,
	ctt_cst,
	ctt_sym,
	ctt_neg,
	ctt_add,
	ctt_sub,
	ctt_mul,
	ctt_rem,
	ctt_max,
	ctt_min,
	ctt_ceild,
	ctt_floord,
	ctt_lod,
};

struct condition_tree {
	condition_tree_type	type;

	int			val;
	var_sym			*var;
	operand			op;
	condition_tree		*left;
	condition_tree		*right;

	condition_tree(condition_tree_type type) :
		type(type), val(-1), var(NULL), left(NULL), right(NULL) {}
	condition_tree(int val) :
	    type(ctt_cst), val(val), var(NULL), left(NULL), right(NULL) {}
	condition_tree(var_sym *var) :
	    type(ctt_sym), val(-1), var(var), left(NULL), right(NULL) {}
	condition_tree(const operand &op) :
	    type(ctt_lod), val(-1), var(NULL), op(op), left(NULL), right(NULL) {}
	condition_tree(condition_tree_type type, condition_tree *left) :
			type(type), val(-1), var(NULL),
			left(left), right(NULL) {
		if (!left || left->type == ctt_invalid) {
			if (left)
				delete left;
			this->left = NULL;
			this->type = ctt_invalid;
		} else if (type == ctt_neg && left->type == ctt_cst) {
			this->val *= left->val;
			this->left = NULL;
			delete left;
			this->type = ctt_cst;
		}
	}
	condition_tree(condition_tree_type type,
			condition_tree *left, condition_tree *right) :
			type(type), val(-1), var(NULL),
			left(left), right(right) {
		if (!left || left->type == ctt_invalid ||
		    !right || right->type == ctt_invalid) {
			if (left)
				delete left;
			if (right)
				delete right;
			this->left = NULL;
			this->right = NULL;
			this->type = ctt_invalid;
		} else if (type == ctt_sub) {
			this->type = ctt_add;
			this->right = new condition_tree(ctt_neg, right);
		} else if (type == ctt_mul &&
			   left->type == ctt_cst && right->type == ctt_cst) {
			this->val = left->val * right->val;
			delete left;
			delete right;
			this->left = NULL;
			this->right = NULL;
			this->type = ctt_cst;
		} else if (type == ctt_mul &&
			   left->type == ctt_cst && right->type == ctt_neg) {
			condition_tree *tmp = right;
			left->val *= -1;
			this->right = tmp->left;
			tmp->left = NULL;
			delete tmp;
		} else if (type == ctt_mul &&
			   left->type == ctt_cst && right->type == ctt_add) {
			this->left = new condition_tree(ctt_mul,
					new condition_tree(left->val),
					right->left);
			this->right = new condition_tree(ctt_mul,
					left, right->right);
			right->left = NULL;
			right->right = NULL;
			delete right;
			this->type = ctt_add;
		} else if (type == ctt_mul &&
			   left->type == ctt_cst && right->type == ctt_mul) {
			this->left = new condition_tree(ctt_mul,
					left, right->left);
			this->right = right->right;
			right->left = NULL;
			right->right = NULL;
			delete right;
		}
	}
	condition_tree(condition_tree_type type,
			int val, var_sym *var,
			condition_tree *left, condition_tree *right) :
			type(type), val(val), var(var),
			left(left), right(right) {
	}

	condition_tree *dup() {
		return new condition_tree(type, val, var,
				left ? left->dup() : NULL,
				right ? right->dup() : NULL);
	}

	void dump(FILE *out = stderr, int indent = 0);
	void eliminate_minmax();
	
	~condition_tree() {
		if (left)
			delete left;
		if (right)
			delete right;
	}
};

condition_tree *construct_condition_tree(operand op, int verbose);
condition_tree *construct_condition_tree_expr(instruction *ins, int verbose);
condition_tree *construct_condition_tree(instruction *ins, int verbose);
