#include <isl/map.h>

#include <isa/yaml.h>
#include <isa/pdg.h>

using namespace pdg;

int main(int argc, char * argv[])
{
    PDG *pdg;
    isl_ctx *ctx = isl_ctx_alloc();
    pdg = PDG::Load(stdin, ctx);

    for (int i = 0; i < pdg->dependences.size(); ++i) {
	pdg::dependence *dep = pdg->dependences[i];
	if (dep->type != pdg::dependence::uninitialized)
	    continue;
	pdg::node* to = dep->to;
	int to_access = dep->to_access ? dep->to_access->nr : -1;
	const char *to_f = to->statement->top_function ?
			   to->statement->top_function->name->s.c_str() : "(none)";
	pdg::IslMap *rel = dep->relation;
	isl_map *map = rel->get_isl_map(ctx);
	printf("%s uninitialized at %d,%d (%s,line %d): ",
		dep->array->name->s.c_str(),
		to->nr, to_access, to_f, to->statement->line
		);
	isl_map_print(map, stdout, 0, ISL_FORMAT_ISL);
	printf("\n");
	isl_map_free(map);
    }

    for (int i = 0; i < pdg->dependences.size(); ++i) {
	pdg::dependence *dep = pdg->dependences[i];
	if (dep->type == pdg::dependence::uninitialized)
	    continue;
	pdg::node* from = dep->from;
	pdg::node* to = dep->to;
	int from_access = dep->from_access ? dep->from_access->nr : -1;
	int to_access = dep->to_access ? dep->to_access->nr : -1;
	const char *from_f = from->statement->top_function ?
			     from->statement->top_function->name->s.c_str() :
			     "(none)";
	const char *to_f = to->statement->top_function ?
			   to->statement->top_function->name->s.c_str() : "(none)";
	pdg::IslMap *rel = dep->relation;
	isl_map *map = rel->get_isl_map(ctx);
	printf("type: %d\n", dep->type);
	printf("%d,%d (%s,line %d) --(",
		from->nr, from_access, from_f, from->statement->line);
	if (dep->array)
		printf("%s", dep->array->name->s.c_str());
	for (int j = 0; j < dep->from_controls.size(); ++j) {
		if (j)
			printf(",");
		printf("%s", dep->from_controls[j]->s.c_str());
	}
	for (int j = 0; j < dep->to_controls.size(); ++j) {
		if (j)
			printf(",");
		else
			printf("; ");
		printf("%s", dep->to_controls[j]->s.c_str());
	}
	printf(")--> %d,%d (%s,line %d): ",
		to->nr, to_access, to_f, to->statement->line);
	isl_map_print(map, stdout, 0, ISL_FORMAT_ISL);
	printf("\n");
	isl_map_free(map);
	if (dep->extended_relation) {
		pdg::IslMap *rel = dep->extended_relation;
		isl_map *map = rel->get_isl_map(ctx);
		printf("extended: ");
		isl_map_print(map, stdout, 0, ISL_FORMAT_ISL);
		printf("\n");
		isl_map_free(map);
	}
	if (dep->controlled_relation) {
		isl_map *map = dep->controlled_relation->map;
		printf("controlled: ");
		isl_map_print(map, stdout, 0, ISL_FORMAT_ISL);
		printf("\n");
	}
    }

    pdg->free();
    delete pdg;
    isl_ctx_free(ctx);

    return 0;
}
