#include "msa.h"

int MSA::add_firing(int s, int t, __isl_take isl_map *map)
{
	int f;
	char name[50];

	snprintf(name, sizeof(name), "s%d", s);
	map = isl_map_set_tuple_name(map, isl_dim_in, name);
	snprintf(name, sizeof(name), "s%d", t);
	map = isl_map_set_tuple_name(map, isl_dim_out, name);
	for (f = 0; f < firing.size(); ++f) {
		if (!isl_map_fast_is_equal(firing[f], map))
			continue;
		isl_map_free(map);
		return f;
	}
	firing.push_back(map);

	return f;
}

/* Return an index to the state of the MSA that corresponds to the
 * given pair of computations, adding this state if necessary.
 * If the state needs to be added, then also add all connections
 * for the state to other states.
 */
int MSA::add(computation *c1, computation *c2)
{
	int s;

	for (s = 0; s < state.size(); ++s)
		if (state[s].first == c1 && state[s].second == c2)
			return s;
	state.push_back(computation_pair(c1, c2));

	if (!c1->is_copy() && !c2->is_copy() &&
	    (strcmp(c1->operation, c2->operation) || c1->arity != c2->arity)) {
		fail_state.push_back(s);
	} else if (c1->is_copy()) {
		for (int i = 0; i < c1->edges.size(); ++i) {
			edge *e1 = c1->edges[i];
			isl_space *dim;
			isl_map *id, *map;
			int t, f;
			assert(e1->type != edge::expansion);
			t = add(e1->source, c2);
			dim = isl_space_map_from_set(isl_set_get_space(c2->domain));
			id = isl_map_identity(dim);
			map = isl_map_product(isl_map_copy(e1->relation), id);
			f = add_firing(s, t, map);
			transition.push_back(transition_t(state_pair(s, t), f));
		}
	} else if (c2->is_copy()) {
		for (int i = 0; i < c2->edges.size(); ++i) {
			edge *e2 = c2->edges[i];
			isl_space *dim;
			isl_map *id, *map;
			int t, f;
			assert(e2->type != edge::expansion);
			t = add(c1, e2->source);
			dim = isl_set_get_space(c1->domain);
			id = isl_map_identity(isl_space_map_from_set(dim));
			map = isl_map_product(id, isl_map_copy(e2->relation));
			f = add_firing(s, t, map);
			transition.push_back(transition_t(state_pair(s, t), f));
		}
	} else if (c1->is_input() && c2->is_input()) {
		if (strcmp(c1->operation, c2->operation))
			fail_state.push_back(s);
		else
			input_state.push_back(s);
	} else if (c1->is_input() || c2->is_input()) {
		fail_state.push_back(s);
	} else if (c1->arity == 0) {
		cst_state.push_back(s);
	} else {
		for (int i = 0; i < c1->edges.size(); ++i) {
			edge *e1 = c1->edges[i];
			assert(e1->type != edge::expansion);
			for (int j = 0; j < c2->edges.size(); ++j) {
				edge *e2 = c2->edges[j];
				assert(e2->type != edge::expansion);
				int t, f;
				isl_map *map;
				if (e1->pos != e2->pos)
					continue;
				t = add(e1->source, e2->source);
				map = isl_map_product(
					isl_map_copy(e1->relation),
					isl_map_copy(e2->relation));
				f = add_firing(s, t, map);
				transition.push_back(
					transition_t(state_pair(s, t), f));
			}
		}
	}

	return s;
}
