#include <isa/pdg.h>
#include <algorithm>
#include <vector>
#include "version.h"

#include <isl/set.h>
#include <isl/map.h>

using namespace pdg;
using std::vector;

static at_init register_history_entry(history_entry::register_type);

void history_entry::register_type()
{
    static struct_description d = { create };
    YAML_PTR_FIELD(d, history_entry, tool, str);
    YAML_PTR_FIELD(d, history_entry, version, str);
    YAML_SEQ_FIELD(d, history_entry, arguments, str);
    structure::register_type("perl/history_entry", &typeid(history_entry), &d.d);
}

static at_init register_pdg(PDG::register_type);

void PDG::register_type()
{
    static struct_description pdg_d = { create };
    YAML_INT_FIELD(pdg_d, PDG, dimension);
    YAML_INT_FIELD(pdg_d, PDG, cacheline);
    YAML_INT_FIELD(pdg_d, PDG, root);
    YAML_PTR_FIELD(pdg_d, PDG, placement, str);
    YAML_SEQ_FIELD(pdg_d, PDG, nodes, node);
    YAML_SEQ_FIELD(pdg_d, PDG, arrays, array);
    YAML_SEQ_FIELD(pdg_d, PDG, dependences, dependence);
    YAML_SEQ_FIELD(pdg_d, PDG, params, parameter);
    YAML_PTR_FIELD(pdg_d, PDG, context, IslSet);
    YAML_SEQ_FIELD(pdg_d, PDG, history, history_entry);
    YAML_PTR_FIELD(pdg_d, PDG, name, str);
    YAML_SEQ_FIELD(pdg_d, PDG, statement_dimensions, int);
    structure::register_type("perl/PDG", &typeid(PDG), &pdg_d.d);
}

PDG *PDG::Load(char *str, isl_ctx *ctx)
{
    return yaml::Load<PDG>(str, ctx);
}

PDG *PDG::Load(FILE *fp, isl_ctx *ctx)
{
    return yaml::Load<PDG>(fp, ctx);
}

PDG::~PDG()
{
    if (ctx_created)
	isl_ctx_free(ctx);
}

isl_ctx *PDG::get_isl_ctx()
{
    if (!ctx) {
	ctx = isl_ctx_alloc();
	assert(ctx);
	ctx_created = true;
    }
    return ctx;
}

isl_set *PDG::get_context_isl_set()
{
    isl_ctx *ctx = get_isl_ctx();
    isl_set *set;

    if (context) {
	set = context->get_isl_set(ctx);
    } else {
	isl_space *dims = isl_space_set_alloc(ctx, params.size(), 0);
	isl_dim_set_parameter_names(dims, params);
	set = isl_set_universe(dims);
    }

    return set;
}

void PDG::add_history_line(const char *tool, int argc, char * argv[])
{
    history_entry *he = new history_entry();
    he->tool = new str(string(tool));
    he->version = new str(string(pdg_version()));
    for (int i = 1; i < argc; ++i)
	he->arguments.push_back(new str(argv[i]));
    history.push_back(he);
}

static at_init register_array(array::register_type);

void array::register_type()
{
    static const char *at_names[array::last];
    at_names[input] = "input";
    at_names[output] = "output";
    at_names[temp] = "temp";
    static struct_description d = { create };
    YAML_INT_FIELD(d, array, id);
    YAML_PTR_FIELD(d, array, name, str);
    YAML_SEQ_FIELD(d, array, dims, int);
    YAML_PTR_FIELD(d, array, element_type, str);
    YAML_ENUM_FIELD(d, array, type, at_names);
    YAML_SEQ_FIELD(d, array, analysis_performed, dependence_type);
    YAML_PTR_FIELD(d, array, value_bounds, IslSet);
    YAML_INT_FIELD(d, array, uniquely_defined);
    structure::register_type("perl/array", &typeid(array), &d.d);
}

static at_init register_parameter(parameter::register_type);

void parameter::register_type()
{
    static struct_description d = { create };
    YAML_INT_FIELD(d, parameter, id);
    YAML_PTR_FIELD(d, parameter, name, str);
    YAML_PTR_FIELD(d, parameter, value, integer);
    structure::register_type("perl/parameter", &typeid(parameter), &d.d);
}

void pdg::isl_dim_set_parameter_names(isl_space *dim,
					const seq<parameter>& params)
{
	for (int i = 0; i < params.size(); ++i)
		dim = isl_space_set_dim_name(dim, isl_dim_param, i,
					params[i]->name->s.c_str());
}

isl_set *pdg::isl_set_set_parameter_names(isl_set *set,
					    const seq<parameter>& params)
{
	for (int i = 0; i < params.size(); ++i)
		set = isl_set_set_dim_name(set, isl_dim_param, i,
					    params[i]->name->s.c_str());
	return set;
}

static at_init register_access(pdg::access::register_type);

void pdg::access::register_type()
{
    static const char *at_names[access::last];
    at_names[write] = "write";
    at_names[read] = "read";
    static struct_description d = { create };
    YAML_PTR_FIELD(d, pdg::access, array, pdg::array);
    YAML_ENUM_FIELD(d, pdg::access, type, at_names);
    YAML_PTR_FIELD(d, pdg::access, map, IslMap);
    YAML_INT_FIELD(d, pdg::access, nr);
    YAML_PTR_FIELD(d, pdg::access, extension, IslMap);
    YAML_PTR_FIELD(d, pdg::access, extended_map, IslMap);
    YAML_SEQ_FIELD(d, pdg::access, nested, call_or_access);
    YAML_SEQ_FIELD(d, pdg::access, sources, IslMap);
    structure::register_type("perl/access", &typeid(pdg::access), &d.d);
}

static at_init register_function_call(function_call::register_type);

void function_call::register_type()
{
    static struct_description s_d = { create };
    YAML_SEQ_FIELD(s_d, function_call, arguments, call_or_access);
    YAML_PTR_FIELD(s_d, function_call, name, str);
    structure::register_type("perl/function_call", &typeid(function_call), &s_d.d);
}

static at_init register_call_or_access(call_or_access::register_type);

void call_or_access::register_type()
{
    static const char *cat_names[call_or_access::last];
    cat_names[t_access] = "access";
    cat_names[t_call] = "call";
    static struct_description s_d = { create };
    YAML_ENUM_FIELD(s_d, pdg::call_or_access, type, cat_names);
    YAML_PTR_FIELD(s_d, pdg::call_or_access, access, pdg::access);
    YAML_PTR_FIELD(s_d, pdg::call_or_access, call, function_call);
    structure::register_type("perl/call_or_access", &typeid(call_or_access), &s_d.d);
}

static at_init register_stat(statement::register_type);

void statement::register_type()
{
    static struct_description s_d = { create };
    YAML_INT_FIELD(s_d, statement, operation);
    YAML_INT_FIELD(s_d, statement, line);
    YAML_SEQ_FIELD(s_d, statement, accesses, access);
    YAML_PTR_FIELD(s_d, statement, top_function, function_call);
    YAML_SEQ_FIELD(s_d, statement, top_outputs, access);
    structure::register_type("perl/statement", &typeid(statement), &s_d.d);
}

static at_init register_node(node::register_type);

void node::register_type()
{
    static struct_description node_d = { create };
    YAML_INT_FIELD(node_d, node, nr);
    YAML_PTR_FIELD(node_d, node, name, str);
    YAML_PTR_FIELD(node_d, node, source, IslSet);
    YAML_PTR_FIELD(node_d, node, schedule, IslMap);
    YAML_PTR_FIELD(node_d, node, statement, pdg::statement);
    YAML_SEQ_FIELD(node_d, node, prefix, int);
    YAML_SEQ_FIELD(node_d, node, filters, call_or_access);
    structure::register_type("perl/node", &typeid(node), &node_d.d);
}

static at_init register_dependence(dependence::register_type);

void dependence::register_type()
{
    static const char *dt_names[dependence::last];
    dt_names[flow] = "flow";
    dt_names[anti] = "anti";
    dt_names[pn] = "pn";
    dt_names[pn_union] = "pn_union";
    dt_names[pn_part] = "pn_part";
    dt_names[pn_wire] = "pn_wire";
    dt_names[pn_hold] = "pn_hold";
    dt_names[pn_shift] = "pn_shift";
    dt_names[reuse] = "reuse";
    dt_names[reuse_pair] = "reuse_pair";
    dt_names[output] = "output";
    dt_names[uninitialized] = "uninitialized";
    static struct_description dependence_d = { create };
    YAML_PTR_FIELD(dependence_d, dependence, relation, IslMap);
    YAML_PTR_FIELD(dependence_d, dependence, extended_relation, IslMap);
    YAML_PTR_FIELD(dependence_d, dependence, controlled_relation, IslMap);
    YAML_PTR_FIELD(dependence_d, dependence, from, pdg::node);
    YAML_PTR_FIELD(dependence_d, dependence, to, pdg::node);
    YAML_PTR_FIELD(dependence_d, dependence, from_access, pdg::access);
    YAML_PTR_FIELD(dependence_d, dependence, to_access, pdg::access);
    YAML_ENUM_FIELD(dependence_d, dependence, type, dt_names);
    YAML_PTR_FIELD(dependence_d, dependence, array, pdg::array);
    YAML_SEQ_FIELD(dependence_d, dependence, from_controls, str);
    YAML_SEQ_FIELD(dependence_d, dependence, to_controls, str);
    YAML_INT_FIELD(dependence_d, dependence, reordering);
    YAML_INT_FIELD(dependence_d, dependence, multiplicity);
    YAML_PTR_FIELD(dependence_d, dependence, size, enumerator);
    YAML_PTR_FIELD(dependence_d, dependence, container, dependence);
    YAML_PTR_FIELD(dependence_d, dependence, value_size, integer);
    structure::register_type("perl/dependence", &typeid(dependence), &dependence_d.d);

    static struct_description dependence_type_d = { dependence_type::create };
    YAML_ENUM_FIELD(dependence_type_d, dependence_type, type, dt_names);
    structure::register_type("perl/dependence_type", &typeid(dependence_type), &dependence_type_d.d);
}

static at_init register_enumerator(pdg::enumerator::register_type);

void pdg::enumerator::register_type()
{
    static struct_description e_d = { create, convert };
    structure::register_type("perl/PDG::enumerator", &typeid(enumerator), &e_d.d);
}

void pdg::enumerator::init(SyckParser *p, SyckNode *n)
{
    s = new string(n->data.str->ptr, n->data.str->len);
}

serialize *pdg::enumerator::convert(anchor_map *am, serialize *node,
	const type_info *type)
{
	str *s = dynamic_cast<str *>(node);
	pdg::enumerator *e;

	if (!s)
		return NULL;

	e = new enumerator(new string(s->s));

	delete node;

	return e;
}

void pdg::enumerator::dump(emitter& e)
{
    if (s)
	yll_emitter_write_string(e.e, s->c_str());
    else if (value >= 0)
	yll_emitter_write_int(e.e, value);
    else
	assert(0);
}

pdg::enumerator::~enumerator()
{
    if (s)
	delete s;
}

static at_init register_matrix(pdg::Matrix::register_type);

void pdg::Matrix::register_type()
{
    static struct_description m_d = { create, convert };
    structure::register_type("perl/PDL::Matrix", &typeid(Matrix), &m_d.d);
}

void pdg::Matrix::init(SyckParser *p, SyckNode *n)
{
    el = vector<vector<int> >(n->data.list->idx);
    for (int i = 0; i < n->data.list->idx; i++) {
	serialize *val;
	SYMID oid = syck_seq_read(n, i);
	syck_lookup_sym( p, oid, (char **)&val );
	seq<serialize> * row = dynamic_cast<seq<serialize>*>(val);
	el[i] = vector<int>(row->size());
	for (int j = 0; j < row->size() ; ++j) {
	    str * s = dynamic_cast<str *>((*row)[j]);
	    el[i][j] = atoi(s->s.c_str());
	    delete s;
	}
	delete val;
    }
}

/* Convert a sequence of a sequence of integers (s) to a Matrix object.
 */
serialize *pdg::Matrix::convert(anchor_map *am, serialize *s,
	const type_info * type)
{
	seq<serialize> *matrix = dynamic_cast<seq<serialize> *>(s);
	pdg::Matrix *M;

	if (!matrix)
		return NULL;

	M = new Matrix();
	M->el = vector<vector<int> >(matrix->size());
	for (int i = 0; i < matrix->size(); ++i) {
		serialize *val = (*matrix)[i];
		seq<serialize> *row = dynamic_cast<seq<serialize> *>(val);
		M->el[i] = vector<int>(row->size());
		for (int j = 0; j < row->size() ; ++j) {
			str *s = dynamic_cast<str *>((*row)[j]);
			M->el[i][j] = atoi(s->s.c_str());
			delete s;
		}
		delete val;
	}

	delete s;

	return M;
}

void pdg::Matrix::dump(emitter& e)
{
    yll_emitter_start_seq(e.e);
    for (int i = 0; i < el.size(); ++i) {
	yll_emitter_set_options(e.e, YLL_INLINE_SEQ);
	yll_emitter_start_seq(e.e);
	for (int j = 0; j < el[i].size(); ++j) {
	    yll_emitter_write_int(e.e, el[i][j]);
	}
	yll_emitter_end_seq(e.e);
    }
    yll_emitter_end_seq(e.e);
}

void pdg::Matrix::dump()
{
    for (int i = 0; i < el.size(); ++i) {
	for (int j = 0; j < el[i].size(); ++j) {
	    printf("%d ", el[i][j]);
	}
	puts("");
    }
}

//static at_init register_hmatrix(HMatrix::register_type);

void HMatrix::register_type()
{
    static struct_description m_d = { create };
    structure::register_type("perl/PDL::HMatrix", &typeid(HMatrix), &m_d.d);
}

static at_init register_isl_set(IslSet::register_type);

void IslSet::register_type()
{
	static struct_description s_d = { create, convert };
	structure::register_type("", &typeid(IslSet), &s_d.d);
}

serialize *IslSet::convert(anchor_map *am, serialize *node,
	const type_info *type)
{
	UnionSet *us;
	dict *d = dynamic_cast<dict *>(node);
	str *s = dynamic_cast<str *>(node);
	isl_ctx *ctx = (isl_ctx *) am->user;
	isl_set *set;

	if (d)
		node = structure::convert(am, node, &typeid(UnionSet));

	us = dynamic_cast<UnionSet *>(node);
	if (us) {
		set = us->get_isl_set(ctx);
		us->params.resize(0);
		us->free();
		delete us;
		return new IslSet(set);
	}
	if (s) {
		set = isl_set_read_from_str(ctx, s->s.c_str());
		delete node;
		return new IslSet(set);
	}

	return NULL;
}

void IslSet::dump(emitter &e)
{
	isl_printer *p;
	char *s;

	if (!set) {
		yll_emitter_write_null(e.e);
		return;
	}

	p = isl_printer_to_str(isl_set_get_ctx(set));
	p = isl_printer_print_set(p, set);
	s = isl_printer_get_str(p);
	isl_printer_free(p);

	yll_emitter_write_string(e.e, s);

	::free(s);
}

static at_init register_isl_map(IslMap::register_type);

void IslMap::register_type()
{
	static struct_description s_d = { create, convert };
	structure::register_type("", &typeid(IslMap), &s_d.d);
}

serialize *IslMap::convert(anchor_map *am, serialize *node,
	const type_info *type)
{
	UnionMap *um;
	dict *d = dynamic_cast<dict *>(node);
	str *s = dynamic_cast<str *>(node);
	isl_ctx *ctx = (isl_ctx *) am->user;
	isl_map *map;

	if (d)
		node = structure::convert(am, node, &typeid(UnionMap));

	um = dynamic_cast<UnionMap *>(node);
	if (um) {
		map = um->get_isl_map(ctx);
		um->params.resize(0);
		um->free();
		delete um;
		return new IslMap(map);
	}
	if (s) {
		map = isl_map_read_from_str(ctx, s->s.c_str());
		delete node;
		return new IslMap(map);
	}

	return NULL;
}

void IslMap::dump(emitter &e)
{
	isl_printer *p;
	char *s;

	if (!map) {
		yll_emitter_write_null(e.e);
		return;
	}

	p = isl_printer_to_str(isl_map_get_ctx(map));
	p = isl_printer_print_map(p, map);
	s = isl_printer_get_str(p);
	isl_printer_free(p);

	yll_emitter_write_string(e.e, s);

	::free(s);
}

static at_init register_set(UnionSet::register_type);

void UnionSet::register_type()
{
    static struct_description s_d = { create };
    YAML_INT_FIELD(s_d, UnionSet, dim);
    YAML_SEQ_FIELD(s_d, UnionSet, constraints, Matrix);
    YAML_SEQ_FIELD(s_d, UnionSet, params, parameter);
    structure::register_type("perl/PDG::UnionSet", &typeid(UnionSet), &s_d.d);
}

void UnionSet::fill_fields(anchor_map *am)
{
    struct_description * sd = (struct_description *)(t_desc())[&typeid(*this)];
    map<const type_info *, description*>& desc = serialize::t_desc();

    map<string,serialize_p>::iterator i;
    for (i = d->begin(); i != d->end(); ) {
	map<string,serialize_p>::iterator j = i++;
	if (sd->m.find(j->first) != sd->m.end()) {
	    serialize *el = j->second;

	    if (sd->m[j->first]->set(am, this, j->second))
		d->erase(j);
	    /* Perhaps it was a Set instead of a UnionSet */
	    else if (j->first == string("constraints")) {
		Matrix *m = dynamic_cast<Matrix *>(j->second);
		if (m) {
		    d->erase(j);
		    constraints.push_back(m);
		}
	    }
	}
    }
}

void UnionSet::dump(emitter& e)
{
    structure::dump(e);
}

isl_basic_map *pdg::Matrix::get_isl_basic_map(isl_space *dim) const
{
	int n_eq = 0;
	isl_ctx *ctx;
	isl_mat *eq;
	isl_mat *ineq;
	int cols;
	int i, j, k, l;
	isl_int v;

	if (el.size() == 0)
		return isl_basic_map_universe(dim);

	cols = el[0].size() - 1;

	for (i = 0; i < el.size(); ++i)
		if (el[i][0] == 0)
			n_eq++;

	ctx = isl_space_get_ctx(dim);
	eq = isl_mat_alloc(ctx, n_eq, cols);
	ineq = isl_mat_alloc(ctx, el.size() - n_eq, cols);

	isl_int_init(v);

	for (k = l = i = 0; i < el.size(); ++i) {
		if (el[i][0] == 0) {
			for (j = 0; j < cols; ++j) {
				isl_int_set_si(v, el[i][1 + j]);
				eq = isl_mat_set_element(eq, k, j, v);
			}
			++k;
		} else {
			for (j = 0; j < cols; ++j) {
				isl_int_set_si(v, el[i][1 + j]);
				ineq = isl_mat_set_element(ineq, l, j, v);
			}
			++l;
		}
	}

	isl_int_clear(v);

	return isl_basic_map_from_constraint_matrices(dim, eq, ineq,
	    isl_dim_in, isl_dim_out, isl_dim_div, isl_dim_param, isl_dim_cst);
}

isl_basic_set *pdg::Matrix::get_isl_basic_set(isl_space *dim) const
{
	isl_basic_map *bmap;

	dim = isl_space_from_range(dim);
	bmap = get_isl_basic_map(dim);
	return isl_basic_map_range(bmap);
}

static void add_constraints(pdg::Matrix *M, int ineq, __isl_take isl_mat *mat)
{
	int r = isl_mat_rows(mat);
	int c = isl_mat_cols(mat);
	isl_int v;

	isl_int_init(v);

	for (int i = 0; i < r; ++i) {
		std::vector<int> row;
		row.push_back(ineq);
		for (int j = 0; j < c; ++j) {
			isl_mat_get_element(mat, i, j, &v);
			row.push_back(isl_int_get_si(v));
		}
		M->el.push_back(row);
	}

	isl_int_clear(v);
	isl_mat_free(mat);
}

static int add_basic_set_to_constraints(__isl_take isl_basic_set *bset,
	void *user)
{
	isl_mat *mat;
	seq<pdg::Matrix> *constraints  = (seq<pdg::Matrix> *) user;

	pdg::Matrix *M = new pdg::Matrix;
	mat = isl_basic_set_equalities_matrix(bset, isl_dim_set,
		isl_dim_div, isl_dim_param, isl_dim_cst);
	add_constraints(M, 0, mat);
	mat = isl_basic_set_inequalities_matrix(bset, isl_dim_set,
		isl_dim_div, isl_dim_param, isl_dim_cst);
	add_constraints(M, 1, mat);

	constraints->push_back(M);
	isl_basic_set_free(bset);
	return 0;
}

UnionSet::UnionSet(isl_set *set, seq<parameter>& p) :
	dim(isl_set_dim(set, isl_dim_set)), params(p)
{
	isl_set_foreach_basic_set(set, add_basic_set_to_constraints, &constraints);
}

isl_set *UnionSet::get_isl_set(isl_ctx *ctx) const
{
	isl_space *dims = isl_space_set_alloc(ctx, params.size(), dim);
	isl_dim_set_parameter_names(dims, params);
	isl_set *set = isl_set_empty(isl_space_copy(dims));

	for (int i = 0; i < constraints.size(); ++i) {
		isl_basic_set *bset;

		bset = constraints[i]->get_isl_basic_set(isl_space_copy(dims));

		set = isl_set_union(set, isl_set_from_basic_set(bset));
	}

	isl_space_free(dims);

	return set;
}

static at_init register_map(UnionMap::register_type);

void UnionMap::register_type()
{
    static struct_description m_d = { create };
    YAML_INT_FIELD(m_d, UnionMap, input);
    YAML_INT_FIELD(m_d, UnionMap, output);
    YAML_SEQ_FIELD(m_d, UnionMap, constraints, Matrix);
    YAML_SEQ_FIELD(m_d, UnionMap, params, parameter);
    structure::register_type("perl/PDG::UnionMap", &typeid(UnionMap), &m_d.d);
}

void UnionMap::fill_fields(anchor_map *am)
{
    struct_description * sd = (struct_description *)(t_desc())[&typeid(*this)];
    map<const type_info *, description*>& desc = serialize::t_desc();

    map<string,serialize_p>::iterator i;
    for (i = d->begin(); i != d->end(); ) {
	map<string,serialize_p>::iterator j = i++;
	if (sd->m.find(j->first) != sd->m.end()) {
	    serialize *el = j->second;

	    if (sd->m[j->first]->set(am, this, j->second))
		d->erase(j);
	    /* Perhaps it was a Map instead of a UnionMap */
	    else if (j->first == string("constraints")) {
		Matrix *m = dynamic_cast<Matrix *>(j->second);
		if (m) {
		    d->erase(j);
		    constraints.push_back(m);
		}
	    }
	}
    }
}

void UnionMap::dump(emitter& e)
{
    structure::dump(e);
}

static int add_basic_map_to_constraints(__isl_take isl_basic_map *bmap,
	void *user)
{
	isl_mat *mat;
	seq<pdg::Matrix> *constraints  = (seq<pdg::Matrix> *) user;

	pdg::Matrix *M = new pdg::Matrix;
	mat = isl_basic_map_equalities_matrix(bmap, isl_dim_in,
		isl_dim_out, isl_dim_div, isl_dim_param, isl_dim_cst);
	add_constraints(M, 0, mat);
	mat = isl_basic_map_inequalities_matrix(bmap, isl_dim_in,
		isl_dim_out, isl_dim_div, isl_dim_param, isl_dim_cst);
	add_constraints(M, 1, mat);

	constraints->push_back(M);
	isl_basic_map_free(bmap);
	return 0;
}

UnionMap::UnionMap(isl_map *map, seq<parameter>& p) :
	input(isl_map_n_in(map)), output(isl_map_n_out(map)), params(p)
{
	isl_map_foreach_basic_map(map, add_basic_map_to_constraints, &constraints);
}

isl_map *UnionMap::get_isl_map(isl_ctx *ctx) const
{
	isl_space *dims = isl_space_alloc(ctx, params.size(), input, output);
	isl_dim_set_parameter_names(dims, params);
	isl_map *map = isl_map_empty(isl_space_copy(dims));

	for (int i = 0; i < constraints.size(); ++i) {
		isl_basic_map *bmap;

		bmap = constraints[i]->get_isl_basic_map(isl_space_copy(dims));

		map = isl_map_union(map, isl_map_from_basic_map(bmap));
	}

	isl_space_free(dims);

	return map;
}

/* Schedule the given dependence relations "map" according to the schedules
 * of the source and target of "dep".
 */
__isl_give isl_map *pdg::schedule_dependence(PDG *pdg, pdg::dependence *dep,
	__isl_take isl_map *map)
{
	isl_map *sr = dep->to->schedule->get_isl_map();
	map = isl_map_apply_range(map, sr);
	if (dep->from) {
		isl_map *sw = dep->from->schedule->get_isl_map();
		map = isl_map_apply_domain(map, sw);
	}

	return map;
}

__isl_give isl_map *pdg::scatter_dependence(PDG *pdg, pdg::dependence *dep)
{
	return schedule_dependence(pdg, dep, dep->relation->get_isl_map());
}

/* Assuming this is a read access, construct a map from the domain
 * of the access relation to the access relation of the corresponding
 * write.  If we are unable to determine the corresponding write, then
 * return the empty map.
 *
 * For example, if the access relation is
 *
 *	S[i] -> A[]
 *
 * and there is a single source of the form
 *
 *	[W[0,i] -> A[]] -> [S[i] -> A[]]
 *
 * then the returned map is
 *
 *	S[i] -> [W[0,i] -> A[]]
 */
__isl_give isl_map *pdg::access::extract_access_map()
{
	isl_map *source_map;
	isl_map *access_map;

	if (sources.size() != 1) {
		isl_space *space;
		space = isl_map_get_space(map->map);
		return isl_map_empty(space);
	}
	access_map = map->get_isl_map();
	access_map = isl_map_reverse(isl_map_domain_map(access_map));
	source_map = sources[0]->get_isl_map();
	source_map = isl_map_reverse(source_map);
	access_map = isl_map_apply_range(access_map, source_map);

	return access_map;
}

/* Return the number of outer loops that are affected by filters.
 *
 * A loop iterator is affected by a filter if it is involved
 * in any of the iteration domain constraints that also involves filters,
 * if any of the filter accesses depends on the loop iterator,
 * or if any of the sources of a filter depends on the loop iterator.
 * If no sources were recorded, then we conservatively assume that
 * all loops are affected by filters.
 */
int node::get_filter_depth()
{
	isl_map *value;
	isl_set *dom;
	int n_in, n_out;

	if (filter_depth >= 0)
		return filter_depth;

	filter_depth = 0;
	value = isl_set_unwrap(source->get_isl_set());
	n_in = isl_map_dim(value, isl_dim_in);
	n_out = isl_map_dim(value, isl_dim_out);

	dom = isl_map_domain(isl_map_copy(value));
	value = isl_map_gist_domain(value, isl_set_copy(dom));
	for (int j = n_in - 1; j >= 0; --j) {
		if (!isl_map_involves_dims(value, isl_dim_in, j, 1))
			continue;
		if (j >= filter_depth)
			filter_depth = j + 1;
		break;
	}

	for (int i = 0; i < n_out; ++i) {
		pdg::call_or_access *coa;

		coa = filters[i];
		assert(coa->type == pdg::call_or_access::t_access);

		if (coa->access->sources.size() == 0) {
			filter_depth = n_in;
			break;
		}

		for (int j = 0; j < coa->access->sources.size(); ++j) {
			isl_map *map;
			map = coa->access->sources[j]->get_isl_map();
			map = isl_map_zip(map);
			map = isl_set_unwrap(isl_map_domain(map));
			map = isl_map_gist_range(map, isl_set_copy(dom));

			for (int k = n_in - 1; k >= 0; --k) {
				if (!isl_map_involves_dims(map,
							    isl_dim_out, k, 1))
					continue;
				if (k >= filter_depth)
					filter_depth = k + 1;
				break;
			}

			isl_map_free(map);
		}

		for (int j = n_in - 1; j >= 0; --j) {
			if (!isl_map_involves_dims(coa->access->map->map,
						    isl_dim_in, j, 1))
				continue;
			if (j >= filter_depth)
				filter_depth = j + 1;
			break;
		}

	}
	isl_set_free(dom);
	isl_map_free(value);

	return filter_depth;
}
