#include <iostream>

#include <isa/yaml.h>
#include <isa/pdg.h>

using namespace std;
using namespace pdg;

int main(int argc, char * argv[])
{
    PDG *pdg;
    isl_ctx *ctx = isl_ctx_alloc();
    pdg = PDG::Load(stdin, ctx);

    printf("digraph dummy {\n");
    for (int i = 0; i < pdg->nodes.size(); ++i) {
	pdg::node *node = pdg->nodes[i];
	if (node->statement->top_function)
	    printf("ND_%d [label = \"%s\"];\n", 
		    node->nr, node->statement->top_function->name->s.c_str());
	else
	    printf("ND_%d;\n", node->nr);
    }
    for (int i = 0; i < pdg->dependences.size(); ++i) {
	pdg::dependence *dep = pdg->dependences[i];
	if (dep->type == pdg::dependence::uninitialized)
	    continue;
	if (dep->type == pdg::dependence::pn_part)
	    continue;
	pdg::node* from = dep->from;
	pdg::node* to = dep->to;
	printf("ND_%d -> ND_%d [label = \"", from->nr, to->nr);
	if (dep->array)
		printf("%s", dep->array->name->s.c_str());
	for (int j = 0; j < dep->from_controls.size(); ++j) {
		if (j)
			printf(",");
		printf("%s", dep->from_controls[j]->s.c_str());
	}
	for (int j = 0; j < dep->to_controls.size(); ++j) {
		if (j)
			printf(",");
		else
			printf("; ");
		printf("%s", dep->to_controls[j]->s.c_str());
	}
	printf(": %s\"%s];\n",
		dep->size ? dep->size->s->c_str() : "?",
		dep->type == pdg::dependence::pn_union ? " color=\"blue\"" :
		dep->type == pdg::dependence::pn_hold ? " color=\"cyan\"" :
		dep->type == pdg::dependence::pn_shift ? " color=\"magenta\"" :
		dep->reordering ? 
		    (dep->multiplicity ? " color=\"yellow\"": " color=\"green\"") : 
		    (dep->multiplicity ? " color=\"red\"": ""));
    }
    printf("}\n");

    pdg->free();
    delete pdg;
    isl_ctx_free(ctx);

    return 0;
}
