#ifndef SIZE_H
#define SIZE_H

#include <isl/polynomial.h>
#include <barvinok/isl.h>
#include <isa/pdg.h>
#include "size_options.h"

namespace size {

struct enumerator {
    enum type {
	fixed,
	qp,
	fold,
    } type;
    int v;
    pdg::PDG *pdg;
    isl_pw_qpolynomial *pwqp;
    isl_pw_qpolynomial_fold *pwf;
    enumerator(int v) : type(fixed), v(v) {}
    enumerator(__isl_take isl_pw_qpolynomial *pwqp, pdg::PDG *pdg) :
	type(qp), pwqp(pwqp), pdg(pdg) {}
    enumerator(__isl_take isl_pw_qpolynomial_fold *pwf, pdg::PDG *pdg) :
	type(fold), pwf(pwf), pdg(pdg) {}
    operator pdg::enumerator *() const;
    /* evaluate the enumerator at the values stored in the parameters */
    integer *evaluate() const;
    ~enumerator() {
	if (type == qp)
	    isl_pw_qpolynomial_free(pwqp);
	if (type == fold)
	    isl_pw_qpolynomial_fold_free(pwf);
    }
};

size::enumerator *selfloop_size(pdg::PDG *pdg, __isl_take isl_map *dep,
			       size_options *options);
isl_pw_qpolynomial *selfloop_tokens_in_fifo(__isl_take isl_map *dep,
	__isl_take isl_set *ref);

}

#endif
