#include <sys/wait.h>
#include <dlfcn.h>
#include <cloog/isl/cloog.h>
#include "size_cloog.h"
#include "clast_scanner.h"

namespace size {

struct channel_counter : public clast_scanner {
    char tmpfile[L_tmpnam];
    char buffer[L_tmpnam+40];

    int count();
    void print_user(clast_user_stmt *s);
    void print_full(CloogNames* names, clast_stmt *s);
};

void channel_counter::print_full(CloogNames* names, clast_stmt *s)
{
    int ret;
    tmpnam(tmpfile);
    snprintf(buffer, 200, "gcc -x c -fPIC -O2 -shared - -o %s", tmpfile);
    dst = popen(buffer, "w");
    assert(dst);
    print_header();
    fprintf(dst, "long __size_count() {\n");
    indent += 4;
    print_iterators(names);
    print_indent();
    fprintf(dst, "long count = 0, max = 0;\n");

    print(s);

    print_indent();
    fprintf(dst, "assert(count == 0);\n");
    fprintf(dst, "return max;\n");
    fprintf(dst, "}\n");
    ret = pclose(dst);
    assert(WEXITSTATUS(ret) == 0);
}

int channel_counter::count() {
    long (*c)();
    void *module;
    long max;
    char *error;

    module = dlopen(tmpfile, RTLD_LAZY);
    unlink(tmpfile);
    if ((error = dlerror()) != NULL) {
	fprintf(stderr, "%s\n", error);
	assert(0);
    }
    assert(module);
    c = (long (*)()) dlsym(module, "__size_count");
    if ((error = dlerror()) != NULL) {
	fprintf(stderr, "%s\n", error);
	assert(0);
    }
    max = c();
    dlclose(module);
    return max;
}

void channel_counter::print_user(clast_user_stmt *u)
{
    print_indent();
    if (u->statement->usr == (void *)1) {
	fprintf(dst, "if (++count > max)\n");
	print_indent();
	fprintf(dst, "    max = count;\n");
    } else
	fprintf(dst, "--count;\n");
}

static int scan_program(clast_stmt *stmt)
{
    int max;

    channel_counter cs;
    cs.print_full(((clast_root *)stmt)->names, stmt);
    max = cs.count();

    return max;
}

int selfloop_size_cloog(__isl_take isl_set *write, __isl_take isl_set *read)
{
    isl_ctx *ctx = isl_set_get_ctx(write);
    isl_space *dim;
    isl_set *universe;
    CloogState *state = cloog_isl_state_malloc(ctx);
    CloogOptions *options = cloog_options_malloc(state);
    CloogDomain *context, *domain;
    CloogUnionDomain *ud;
    CloogInput *input;
    struct clast_stmt *stmt;

    options->compilable = 1;

    fprintf(stderr, "sizecloog\n");

    dim = isl_space_set_alloc(ctx, 0, 0);
    universe = isl_set_universe(dim);
    context = cloog_domain_from_isl_set(universe);

    ud = cloog_union_domain_alloc(0);
    domain = cloog_domain_from_isl_set(read);
    ud = cloog_union_domain_add_domain(ud, NULL, domain, NULL, NULL);
    domain = cloog_domain_from_isl_set(write);
    ud = cloog_union_domain_add_domain(ud, NULL, domain, NULL, (void *)1);

    input = cloog_input_alloc(context, ud);
    stmt = cloog_clast_create_from_input(input, options);

    int size = scan_program(stmt);
    cloog_clast_free(stmt);
    cloog_options_free(options);
    cloog_state_free(state);
    return size;
}

}
