#include <limits.h>
#include <isa/adg.h>
#include "adg_parse.h"
#include "testlib.h"

adg *load_adg(isl_ctx *ctx, const char *base)
{
	char path[PATH_MAX];
	char *srcdir;
	int n;
	FILE *in;
	adg *adg;

	srcdir = getenv("srcdir");
	assert(srcdir);
	n = snprintf(path, sizeof(path), "%s/inputs/%s_pn_adg.yaml",
					srcdir, base);
	assert(n < sizeof(path));
	in = fopen(path, "r");
	assert(in);

	adg = adg_parse(ctx, in);
	assert(adg);
	fclose(in);

	return adg;
}

static void check(isl_ctx *ctx, const char *base)
{
	adg *adg1, *adg2;

	fprintf(stdout, "%s\n", base);

	construct_pdg(base);
	construct_pn(base);
	adg1 = get_adg(ctx, base);
	adg2 = load_adg(ctx, base);
	assert(adg1);
	assert(adg2);
	isl_id_free(adg1->name);
	adg1->name = isl_id_copy(adg2->name);
	assert(adg1->is_equal(adg2));
	delete adg1;
	delete adg2;
}

int main(int argc, char *argv[])
{
	isl_ctx *ctx;

	ctx = isl_ctx_alloc();

	check(ctx, "disconnected");
	check(ctx, "wdp");
	check(ctx, "multi_assignment");
	check(ctx, "multi_reuse");
	check(ctx, "odd_even");
	check(ctx, "test_mod");
	check(ctx, "union");
	check(ctx, "wire2");
	check(ctx, "wlan_dec");
	check(ctx, "lsod");
	check(ctx, "sphinx");

	isl_ctx_free(ctx);

	return 0;
}
