#include <assert.h>
#include <limits.h>
#include <stdlib.h>
#ifdef WIN32
#include <windows.h>
#endif
#include "testlib.h"
#include "adg_parse.h"

#include "config.h"

/* For MinGW */
#ifndef WEXITSTATUS
#define WEXITSTATUS(status)	status
#define WIFEXITED(status)	1
#endif

#ifdef HAVE_PET
#define C2PDG "c2pdg"
#else
#define C2PDG "c2pdg_suif"
#endif

using pdg::PDG;

static char *tmp_dir = NULL;

#ifdef WIN32
static char *create_temp_dir()
{
    static char temp_name[MAX_PATH];

    GetTempFileName(".", "isa_test_", 0, temp_name);
    DeleteFile(temp_name);
    CreateDirectory(temp_name, NULL);
    return strdup(temp_name);
}
#else
static char *create_temp_dir()
{
    return mkdtemp(strdup("isa_test_XXXXXX"));
}
#endif

void remove_tmp_dir(void)
{
    char command[2*PATH_MAX+100];
    int n, ret;

    n = snprintf(command, sizeof(command), "rm -r %s", tmp_dir);
    assert(n < sizeof(command));
    ret = system(command);
    assert(WEXITSTATUS(ret) == 0);

    free(tmp_dir);
}

void construct_pdg(const char *base, const char *options, const char *dir)
{
    char command[2*PATH_MAX+100];
    char *srcdir;
    int n, ret;

    srcdir = getenv("srcdir");
    assert(srcdir);

    if (!tmp_dir) {
	tmp_dir = create_temp_dir();
	assert(tmp_dir);
	atexit(remove_tmp_dir);
    }

    n = snprintf(command, sizeof(command), "cp %s/inputs/%s/%s.c %s",
			srcdir, dir ? dir : "", base, tmp_dir);
    assert(n < sizeof(command));
    ret = system(command);
    assert(WEXITSTATUS(ret) == 0);

    n = snprintf(command, sizeof(command), "../"C2PDG" %s %s/%s.c",
		 options ? options : "", tmp_dir, base);
    assert(n < sizeof(command));
    ret = system(command);
    assert(WEXITSTATUS(ret) == 0);
}

pdg::PDG *get_pdg(isl_ctx *ctx, const char *base, const char *options)
{
    char command[2*PATH_MAX+100];
    int n;
    construct_pdg(base, options);

    FILE *in;
    n = snprintf(command, sizeof(command), "%s/%s.yaml", tmp_dir, base);
    assert(n < sizeof(command));
    in = fopen(command, "r");
    assert(in);

    PDG *pdg;
    pdg = PDG::Load(in, ctx);
    assert(pdg);
    fclose(in);

    return pdg;
}

void construct_pn(const char *base, const char *options)
{
    char command[2*PATH_MAX+100];
    int n, ret;

    n = snprintf(command, sizeof(command), "../pn %s --input %s/%s.yaml",
					    options ? options : "",
					    tmp_dir, base);
    assert(n < sizeof(command));
    ret = system(command);
    assert(WEXITSTATUS(ret) == 0);
}

pdg::PDG *get_pn(isl_ctx *ctx, const char *base, const char *options)
{
    char command[2*PATH_MAX+100];
    int n;

    construct_pn(base, options);
    FILE *in;
    n = snprintf(command, sizeof(command), "%s/%s_pn.yaml", tmp_dir, base);
    assert(n < sizeof(command));
    in = fopen(command, "r");
    assert(in);

    PDG *pdg;
    pdg = PDG::Load(in, ctx);
    assert(pdg);
    fclose(in);

    return pdg;
}

void construct_da(const char *base, const char *options)
{
    char command[2*PATH_MAX+100];
    int n, ret;

    n = snprintf(command, sizeof(command), "../da %s --input %s/%s.yaml",
					    options ? options : "",
					    tmp_dir, base);
    assert(n < sizeof(command));
    ret = system(command);
    assert(WEXITSTATUS(ret) == 0);
}

pdg::PDG *get_da(isl_ctx *ctx, const char *base)
{
    char command[2*PATH_MAX+100];
    int n;

    construct_da(base);
    FILE *in;
    n = snprintf(command, sizeof(command), "%s/%s_da.yaml", tmp_dir, base);
    assert(n < sizeof(command));
    in = fopen(command, "r");
    assert(in);

    PDG *pdg;
    pdg = PDG::Load(in, ctx);
    assert(pdg);
    fclose(in);

    return pdg;
}

void check_eqv(const char *test, const char *options, const char *output)
{
    	char command[2*PATH_MAX+100];
	char *srcdir;
	char dir[PATH_MAX];
	int n, ret;

	srcdir = getenv("srcdir");
	assert(srcdir);

	n = snprintf(dir, sizeof(dir), "eqv/%s", test);
	assert(n < sizeof(dir));

    	construct_pdg("1", NULL, dir);
	construct_da("1", "-t flow,output -A");

    	construct_pdg("2", NULL, dir);
	construct_da("2", "-t flow,output -A");

	n = snprintf(command, sizeof(command),
	    "../eqv %s %s/1_da.yaml %s/2_da.yaml > %s/output",
	    options, tmp_dir, tmp_dir, tmp_dir);
	assert(n < sizeof(command));
	ret = system(command);
	assert(WEXITSTATUS(ret) == 0);

	n = snprintf(command, sizeof(command),
	    "../eqv_cmp %s/inputs/%s/%s %s/output",
	    srcdir, dir, output, tmp_dir);
	assert(n < sizeof(command));
	ret = system(command);
	assert(WEXITSTATUS(ret) == 0);
}

void construct_adg(const char *base)
{
	char command[2 * PATH_MAX + 100];
	int n, ret;

	n = snprintf(command, sizeof(command),
		"../pn2adg -i %s/%s_pn.yaml", tmp_dir, base);
	assert(n < sizeof(command));
	ret = system(command);
	assert(WEXITSTATUS(ret) == 0);
}

adg *get_adg(isl_ctx *ctx, const char *base)
{
	FILE *in;
	char path[PATH_MAX];
	int n;
	adg *adg;

	construct_adg(base);
	n = snprintf(path, sizeof(path), "%s/%s_pn_adg.yaml", tmp_dir, base);
	assert(n < sizeof(path));
	in = fopen(path, "r");
	assert(in);

	adg = adg_parse(ctx, in);
	assert(adg);
	fclose(in);

	return adg;
}
