#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <string>

extern "C" {
#define st_delete STUPID_PROTOTYPING_st_delete
#define st_insert STUPID_PROTOTYPING_st_insert
#include <syck.h>
#include <isa/yll.h>
#undef st_delete
#undef st_insert
int st_delete( register st_table *, register char **, char **value );
int st_insert( register st_table *, register char *, char * );
}

#ifndef HAVE_STRNDUP
extern "C" char *strndup(const char *s, size_t n);
#endif

#include <isa/yaml.h>

using std::string;
using std::vector;

namespace yaml
{

bool emitter::anchor_handled(serialize *s)
{
    if (anchors.find(s) != anchors.end()) {
	yll_emitter_alias(e, anchors[s]);
	return true;
    }
    if (refs[s] > 1)
	anchors[s] = yll_emitter_anchor(e);
    return false;
}

void serialize::collect_references(refset& refs)
{
    refs[this] = 1;
}

void serialize::free()
{
    refset references;
    collect_references(references);
    refset::iterator i;
    for (i = references.begin(); i != references.end(); ++i) {
	if (i->first != this)
	    delete i->first;
    }
}

SyckNode* serialize::bad_anchor_handler(SyckParser *p, char *a)
{
    SyckNode * n = syck_new_str(a, scalar_none);
    serialize_p s = new bad_anchor();

    anchor_map *anchors = (anchor_map *) p->bonus;

    n->id = syck_add_sym(p, (char *)s);

    anchors->add(s, n->id);

    return n;
}

// We may need to replace inside the replacers as well XXX
serialize * anchor_map::replace(SYMID id, serialize *replacement)
{
    serialize *orig;
    st_delete( p->syms, (char **)&id, (char **)&orig );
    assert(anchors[orig].first == id);

    vector<replacer_base *> & replacements = anchors[orig].second;
    anchors[replacement].first = id;
    vector<replacer_base *> & new_replacements = anchors[replacement].second;
    for (int i=0; i < replacements.size(); ++i) {
	replacements[i]->with(replacement);
	new_replacements.push_back(replacements[i]);
    }
    anchors.erase(orig);
    st_insert( p->syms, (char *)id, (char *)replacement );
    return orig;
}

anchor_map::~anchor_map()
{
    map<serialize *, rep_pair >::iterator i;
    for (i = anchors.begin(); i != anchors.end(); ++i) {
	for (int j=0; j < i->second.second.size(); ++j) {
	    delete i->second.second[j];
	}
    }
}

SYMID serialize::syck_handler(SyckParser *p, SyckNode *n)
{
    serialize_p s = 0;
    anchor_map *anchors = (anchor_map *) p->bonus;

    if (n->type_id) {
	if (s_desc().find(string(n->type_id)) != s_desc().end())
	    s = (s_desc())[n->type_id]->create(anchors->user);
    } 
    if (!s) {
	switch (n->kind) {
	case syck_str_kind:
	    s = str::create(anchors->user);
	    break;
	case syck_map_kind:
	    s = dict::create();
	    break;
	case syck_seq_kind:
	    s = seq<serialize>::create();
	    break;
	default:
	    assert(0);
	}
    }

    SYMID id = (n->id == 0) ? syck_add_sym(p, (char *)s) : 0;

    if (id && n->anchor) {
	anchors->add(s, id);
    }

    if (s)
	s->init(p,n);

    // bad anchor
    if (n->id != 0) { 
	serialize *anchor = anchors->replace(n->id, s);
	delete anchor;
	return n->id;
    }

    return id;
}

serialize *serialize::Load(SyckParser *parser, const type_info *type,
				void *user)
{
    SYMID v;
    serialize_p y = NULL;

    anchor_map anchors(parser, user);
    parser->bonus = &anchors;
    syck_parser_handler(parser, syck_handler);
    syck_parser_bad_anchor_handler(parser, bad_anchor_handler);
    syck_parser_implicit_typing( parser, 1 );
    v = syck_parse(parser);
    syck_lookup_sym(parser, v, (char **)&y);

    if (type && y) {
	if (*type != typeid(*y)) {
	    description *desc = t_desc()[type];
	    assert(desc);
	    y = desc->convert(&anchors, y, type);
	}
    }

    syck_free_parser(parser);

    return y;
}

serialize_p serialize::Load(char *str, const type_info *type, void *user)
{
    SyckParser *parser = syck_new_parser();
    syck_parser_str_auto(parser, str, NULL);
    return Load(parser, type, user);
}

serialize_p Load(char *str, const type_info *type, void *user)
{
    return serialize::Load(str, type, user);
}

serialize_p serialize::Load(FILE *fp, const type_info *type, void *user)
{
    SyckParser *parser = syck_new_parser();
    syck_parser_file(parser, fp, NULL);
    return Load(parser, type, user);
}

serialize_p Load(FILE *fp, const type_info *type, void *user)
{
    return serialize::Load(fp, type, user);
}

void serialize::syck_output_handler(YllEmitter *p, char *s, long len)
{
}

void serialize::Dump(FILE *out)
{
    emitter e;
    e.e->out = out;
    collect_references(e.refs);
    yll_emitter_handler(e.e, syck_output_handler);
    Dump(e);
}

void serialize::register_type(const char * type, const type_info *ti, 
			      description * c)
{
    s_desc()[type] = c;
    t_desc()[ti] = c;
}

void str::register_type()
{
    static description str_d = { create, NULL, &typeid(str) };
    serialize::register_type("str", &typeid(str), &str_d);
    serialize::register_type("int", &typeid(str), &str_d);
}

serialize *str::create(void *user)
{
    return new str;
}

void str::init(SyckParser *p, SyckNode *n)
{
    s = string(n->data.str->ptr, n->data.str->len);
}

at_init register_str(str::register_type);

void integer::register_type()
{
    static description integer_d = { create, NULL, &typeid(integer) };
    integer_d.convert = convert;
    serialize::register_type("integer", &typeid(integer), &integer_d);
}

serialize *integer::create(void *user)
{
    return new integer;
}

void integer::init(SyckParser *p, SyckNode *n)
{
    char *s = strndup(n->data.str->ptr, n->data.str->len);
    v = atoi(s);
    ::free(s);
}

serialize_p integer::convert (anchor_map * am, serialize * s, const type_info *type)
{
    str *st = dynamic_cast<str *>(s);
    integer *i = NULL;

    if (st) {
	i = new integer;
	i->v = atoi(st->s.c_str());
    }

    delete s;

    return i;
}

at_init register_integer(integer::register_type);

serialize_p dict::create()
{
    dict *d = new dict;
    return d;
}

static void dict_init(sermap& d, SyckParser *p, SyckNode *n)
{
    anchor_map *anchors = (anchor_map *) p->bonus;
    for (int i = 0; i < n->data.pairs->idx; i++) {
	str *key;
	serialize *val;
	SYMID oidk = syck_map_read( n, map_key, i );
	syck_lookup_sym( p, oidk, (char **)&key );
	SYMID oidv = syck_map_read( n, map_value, i );
	syck_lookup_sym( p, oidv, (char **)&val );
	d[key->s] = val;
	delete key;
    }
    // In a second iteration, since we wouldn't want the
    // memory locations to change for the replacer
    map<string,serialize_p>::iterator i;
    for (i = d.begin(); i != d.end(); ++i) {
	if (anchors->is_anchor(i->second)) {
	    anchors->replace(i->second, new replacer<serialize>(i->second));
	}
    }
}

void dict::init(SyckParser *p, SyckNode *n)
{
    d = new map<string,serialize_p>;
    dict_init(*d, p, n);
}

static void dict_collect(sermap& d, refset& refs)
{
    map<string,serialize_p>::iterator i;
    for (i = d.begin(); i != d.end(); ++i) {
	serialize *el = i->second;
	if (refs.find(el) == refs.end())
	    el->collect_references(refs);
	else
	    ++refs[el];
    }
}

void dict::collect_references(refset& refs)
{
    serialize::collect_references(refs);
    dict_collect(*d, refs);
}

static void dict_dump(sermap& d, emitter& e)
{
    map<string,serialize_p>::iterator i;
    for (i = d.begin(); i != d.end(); ++i) {
	yll_emitter_write_string(e.e, i->first.c_str());
	if (i->second)
	    i->second->Dump(e);
	else
	    yll_emitter_write_null(e.e);
    }
}

void dict::dump(emitter& e)
{
    yll_emitter_start_map(e.e);
    dict_dump(*d, e);
    yll_emitter_end_map(e.e);
}

void structure::init(SyckParser *p, SyckNode *n)
{
    switch (n->kind) {
    case syck_map_kind:
	d = new map<string,serialize_p>;
	dict_init(*d, p, n);
	anchor_map *anchors = (anchor_map *) p->bonus;
	fill_fields(anchors);
	break;
    }
}

void structure::fill_fields(anchor_map *am)
{
    struct_description * sd = (struct_description *)(t_desc())[&typeid(*this)];

    map<string,serialize_p>::iterator i;
    for (i = d->begin(); i != d->end(); ) {
	map<string,serialize_p>::iterator j = i++;
	if (sd->m.find(j->first) != sd->m.end()) {
	    serialize *el = j->second;

	    if (sd->m[j->first]->set(am, this, j->second))
		d->erase(j);
	}
    }
}

serialize_p structure::convert (anchor_map * am, 
				serialize * s, const type_info *type)
{
    dict * d = dynamic_cast<dict *>(s);
    structure * c = NULL;

    if (d) {
	c = dynamic_cast<structure *>(t_desc()[type]->create(am->user));
	assert(t_desc().find(type) != t_desc().end());

	if (am->is_anchor(s)) {
	    serialize *anchor = am->replace(s, c);
	    assert(anchor == s);
	}

	c->d = d->d;
	d->d = NULL;

	c->fill_fields(am);
    }

    delete s;

    return c;
}

void structure::collect_references(refset& refs)
{
    serialize::collect_references(refs);
    if (d)
	dict_collect(*d, refs);
    struct_description * sd = (struct_description *)(t_desc())[&typeid(*this)];
    fieldmap::iterator j;
    for (j = sd->m.begin(); j != sd->m.end(); ++j) {
	sd->m[j->first]->collect(this, refs);
    }
}

void structure::dump(emitter& e)
{
    yll_emitter_start_map(e.e);

    struct_description * sd = (struct_description *)(t_desc())[&typeid(*this)];
    fieldmap::iterator j;
    for (j = sd->m.begin(); j != sd->m.end(); ++j) {
	if (sd->m[j->first]->is_empty(this))
	    continue;

	yll_emitter_write_string(e.e, j->first.c_str());
	sd->m[j->first]->dump(this, e);
    }
    if (d)
	dict_dump(*d, e);

    yll_emitter_end_map(e.e);
}

void structure::register_type(const char * type, const type_info *ti, 
			      description * c)
{
    if (!c->convert)
	c->convert = convert;
    c->type = ti;
    serialize::register_type(type, ti, c);
}

bool ptr_field::set(anchor_map *am, structure * c, serialize *& el)
{
    map<const type_info *, description*>& desc = serialize::t_desc();
    serialize * new_el = el;

    if (*desc[type]->type != typeid(*el))
	new_el = desc[type]->convert(am, el, type);

    if (new_el)  {
	c->*pos = new_el;
	if (am->is_anchor(el)) {
	    am->replace(el, new replacer_base(el),
			    new replacer<serialize>((c->*pos)));
	}
    }

    return el != NULL;
}

void ptr_field::collect(structure * c, refset& refs)
{
    serialize *el = c->*pos;
    if (!el)
	return;

    if (refs.find(el) == refs.end())
	el->collect_references(refs);
    else
	++refs[el];
}

void ptr_field::dump(structure * c, emitter& e)
{
    serialize *el = c->*pos;
    if (el)
	el->Dump(e);
    else
	yll_emitter_write_null(e.e);
}

bool int_field::set(anchor_map *am, structure * c, serialize *& el)
{
    c->*pos = atoi(((str*)el)->s.c_str());
    delete el;

    return true;
}

void int_field::collect(structure * c, refset& refs)
{
}

void int_field::dump(structure * c, emitter& e)
{
    yll_emitter_write_int(e.e, c->*pos);
}

}
